@@ -1,23 +1,17 @@
 import queue
-import re
 import webbrowser
 import requests
-import importlib.resources as pkg_resources
-from typing import Callable
+import json
+from typing import Callable, cast
 from pathlib import Path
 from urllib.parse import urljoin
 from threading import Thread, Event
+from inspect import isgeneratorfunction
 from cheroot import wsgi
 from bottle import Bottle, SimpleTemplate, abort, request, static_file, response, redirect
 
-from browser_ui.utils import Serializable, SerializableCallable, EventType
-from .utils import get_caller_file_abs_path
-
-HEAD_RE = re.compile(r"(<\s*head\b[^>]*>)", re.IGNORECASE)
-BODY_RE = re.compile(r"(<\s*body\b[^>]*>)", re.IGNORECASE)
-INJECTED_SCRIPT_PATH = pkg_resources.files("browser_ui").joinpath("injected_script.js")
-with open(str(INJECTED_SCRIPT_PATH), "r") as f:
-    INJECTED_SCRIPT = f.read()
+from .types import Serializable, SerializableCallable, GeneratorCallable, EventType
+from .utils import inject_script, get_caller_file_abs_path
 
 def server_factory(app: Bottle, port: int, server_name: str='browser-ui-server') -> wsgi.Server:
     return wsgi.Server(
@@ -59,17 +53,23 @@ class BrowserUI:
         self._thread = Thread(target=self._run)
 
         self._app = Bottle()
-        self._method_map: dict[str, SerializableCallable] = {}
+        self._method_map: dict[str, SerializableCallable | GeneratorCallable] = {}
         self._event_map: dict[EventType, list[SerializableCallable]] = {}
         self._format_map: dict[str, Serializable] = {}
         self._server = server_factory(self._app, port)
-        self._sse_queue = queue.Queue()
+        self._sse_queue: queue.Queue[tuple[str, Serializable]] = queue.Queue()
         self._app.route("/", callback=self._serve_static_file)
         self._app.route("/<path:path>", callback=self._serve_static_file)
         self._app.route("/__method__/<method_name>", method="POST", callback=self._serve_method)
         self._app.route("/__event__/<event_name>", method="POST", callback=self._serve_event)
         self._app.route("/__sse__", callback=self._serve_sse)
 
+        @self._app.hook('after_request')  
+        def _set_no_cache():
+            response.set_header('Cache-Control', 'no-cache, no-store, must-revalidate')  
+            response.set_header('Pragma', 'no-cache')  
+            response.set_header('Expires', '0')
+
     @staticmethod
     def _resolve_static_dir_path(static_dir: str) -> Path:
         if Path(static_dir).is_absolute():
@@ -77,15 +77,6 @@ class BrowserUI:
         else:
             return Path(get_caller_file_abs_path(1)).parent.joinpath(static_dir)
 
-    @staticmethod
-    def _inject_script(html: str) -> str:
-        SCRIPT = f"<script>{INJECTED_SCRIPT}</script>"
-        if HEAD_RE.search(html):
-            return HEAD_RE.sub(r"\1" + SCRIPT, html, 1)
-        if BODY_RE.search(html):
-            return BODY_RE.sub(r"\1" + SCRIPT, html, 1)
-        return SCRIPT + html
-
     def _run(self):
         self._server.prepare()
         self._server.serve()
@@ -99,7 +90,7 @@ class BrowserUI:
             assert self._static_dir is not None
             with open(str(self._static_dir.joinpath(path)), "r") as f:
                 html_content = f.read()
-        html_content = self._inject_script(html_content)
+        html_content = inject_script(html_content)
         template = SimpleTemplate(html_content)
         return template.render(self._format_map)
 
@@ -117,9 +108,17 @@ class BrowserUI:
         data = request.json
         if method_name not in self._method_map:
             abort(404, f"Method {method_name} is not implemented.")
-        res = self._method_map[method_name](data)
-        return res
-    
+
+        method = self._method_map[method_name]
+
+        if isgeneratorfunction(method):
+            response.set_header("X-BrowserUI-Stream-Response", "true")
+            for res in cast(GeneratorCallable, method)(data):
+                yield json.dumps(res)
+        else:
+            res = cast(SerializableCallable, method)(data)
+            return json.dumps(res)
+
     def _serve_event(self, event_name: str):
         event = EventType.from_str(event_name)
         if event not in self._event_map:
@@ -133,7 +132,7 @@ class BrowserUI:
         while not self._stop_event.is_set():
             try:
                 event, data = self._sse_queue.get(timeout=0.01)
-                yield f"event: {event}\ndata: {data}\n\n"
+                yield f"event: {event}\ndata: {json.dumps(data)}\n\n"
             except queue.Empty: continue
 
     def add_event_listener(self, event_type: EventType, callback: Callable):
@@ -141,21 +140,24 @@ class BrowserUI:
             self._event_map[event_type] = []
         self._event_map[event_type].append(callback)
 
-    def register(self, method_name: str, method: SerializableCallable):
+    def register_method(self, method_name: str, method: SerializableCallable | GeneratorCallable):
         self._method_map[method_name] = method
 
-    def resgiter_format(self, **args: Serializable):
+    def register_template_vars(self, **args: Serializable):
         for k, v in args.items():
             self._format_map[k] = v
 
-    def send_event(self, event: str, data: str):
+    def send_event(self, event: str, data: Serializable):
         self._sse_queue.put((event, data))
 
-    def start(self):
+    def start(self, path: str | None = None):
         if self._is_used:
             raise RuntimeError("This BrowserUI instance has already been used and cannot be reused.")
         self._thread.start()
-        webbrowser.open_new_tab(f"http://localhost:{self._port}")
+        final_path = f"http://localhost:{self._port}"\
+                     if path is None else\
+                     urljoin(f"http://localhost:{self._port}", path)
+        webbrowser.open_new_tab(final_path)
 
     def stop(self):
         self._stop_event.set()
