@@ -1,2 +1,2 @@
 from .BrowserUI import BrowserUI
-from .utils import EventType
\ No newline at end of file
+from .types import EventType
\ No newline at end of file
